@@ -272,21 +272,20 @@ grub_biosdisk_get_drive (const char *name)
 }
 
 static int
-grub_biosdisk_call_hook (grub_disk_dev_iterate_hook_t hook, void *hook_data,
-			 int drive)
+grub_biosdisk_call_hook (int (*hook) (const char *name), int drive)
 {
   char name[10];
 
   if (cd_drive && drive == cd_drive)
-    return hook ("cd", hook_data);
+    return hook ("cd");
 
   grub_snprintf (name, sizeof (name),
 		 (drive & 0x80) ? "hd%d" : "fd%d", drive & (~0x80));
-  return hook (name, hook_data);
+  return hook (name);
 }
 
 static int
-grub_biosdisk_iterate (grub_disk_dev_iterate_hook_t hook, void *hook_data,
+grub_biosdisk_iterate (int (*hook) (const char *name),
 		       grub_disk_pull_t pull __attribute__ ((unused)))
 {
   int num_floppies;
@@ -305,7 +304,7 @@ grub_biosdisk_iterate (grub_disk_dev_iterate_hook_t hook, void *hook_data,
 	      break;
 	    }
 
-	  if (grub_biosdisk_call_hook (hook, hook_data, drive))
+	  if (grub_biosdisk_call_hook (hook, drive))
 	    return 1;
 	}
       return 0;
@@ -313,14 +312,14 @@ grub_biosdisk_iterate (grub_disk_dev_iterate_hook_t hook, void *hook_data,
     case GRUB_DISK_PULL_REMOVABLE:
       if (cd_drive)
 	{
-	  if (grub_biosdisk_call_hook (hook, hook_data, cd_drive))
+	  if (grub_biosdisk_call_hook (hook, cd_drive))
 	    return 1;
 	}
 
       /* For floppy disks, we can get the number safely.  */
       num_floppies = grub_biosdisk_get_num_floppies ();
       for (drive = 0; drive < num_floppies; drive++)
-	if (grub_biosdisk_call_hook (hook, hook_data, drive))
+	if (grub_biosdisk_call_hook (hook, drive))
 	  return 1;
       return 0;
     default:
