@@ -29,15 +29,11 @@
 struct term_state
 {
   struct term_state *next;
-  struct grub_unicode_glyph *backlog_glyphs;
+  const struct grub_unicode_glyph *backlog_glyphs;
   const grub_uint32_t *backlog_ucs4;
   int backlog_fixed_tab;
   grub_size_t backlog_len;
 
-  int bidi_stack_depth;
-  grub_uint8_t bidi_stack[GRUB_BIDI_MAX_EXPLICIT_LEVEL];
-  int invalid_pushes;
-
   void *free;
   int num_lines;
   char *term_name;
@@ -48,10 +44,7 @@ print_ucs4_real (const grub_uint32_t * str,
 		 const grub_uint32_t * last_position,
 		 int margin_left, int margin_right,
 		 struct grub_term_output *term, int backlog,
-		 int dry_run, int fixed_tab, unsigned skip_lines,
-		 unsigned max_lines,
-		 grub_uint32_t contchar, int fill_right,
-		 struct grub_term_pos *pos);
+		 int dry_run, int fixed_tab);
 
 static struct term_state *term_states = NULL;
 
@@ -133,7 +126,6 @@ grub_set_more (int onoff)
 
 enum
   {
-    GRUB_CP437_UPDOWNARROW     = 0x12,
     GRUB_CP437_UPARROW         = 0x18,
     GRUB_CP437_DOWNARROW       = 0x19,
     GRUB_CP437_RIGHTARROW      = 0x1a,
@@ -161,8 +153,6 @@ map_code (grub_uint32_t in, struct grub_term_output *term)
 	  return GRUB_CP437_LEFTARROW;
 	case GRUB_UNICODE_UPARROW:
 	  return GRUB_CP437_UPARROW;
-	case GRUB_UNICODE_UPDOWNARROW:
-	  return GRUB_CP437_UPDOWNARROW;
 	case GRUB_UNICODE_RIGHTARROW:
 	  return GRUB_CP437_RIGHTARROW;
 	case GRUB_UNICODE_DOWNARROW:
@@ -232,6 +222,7 @@ grub_puts_terminal (const char *str, struct grub_term_output *term)
 	      .variant = 0,
 	      .attributes = 0,
 	      .ncomb = 0,
+	      .combining = 0,
 	      .estimated_width = 1,
 	      .base = *str
 	    };
@@ -252,8 +243,7 @@ grub_puts_terminal (const char *str, struct grub_term_output *term)
       return;
     }
 
-  print_ucs4_real (unicode_str, unicode_last_position, 0, 0, term,
-		   0, 0, 0, 0, -1, 0, 0, 0);
+  print_ucs4_real (unicode_str, unicode_last_position, 0, 0, term, 0, 0, 0);
   grub_free (unicode_str);
 }
 
@@ -392,7 +382,7 @@ read_terminal_list (const char *prefix)
 	}
       
       cur->name = grub_strdup (name);
-      if (! cur->name)
+      if (! name)
 	{
 	  grub_errno = GRUB_ERR_NONE;
 	  grub_free (cur);
@@ -425,6 +415,7 @@ putglyph (const struct grub_unicode_glyph *c, struct grub_term_output *term,
       .variant = 0,
       .attributes = 0,
       .ncomb = 0,
+      .combining = 0,
       .estimated_width = 1
     };
 
@@ -477,7 +468,7 @@ putglyph (const struct grub_unicode_glyph *c, struct grub_term_output *term,
 		}
 	    }
 	  else
-	    code = grub_unicode_get_comb (c) [i].code;
+	    code = c->combining[i].code;
 
 	  grub_ucs4_to_utf8 (&code, 1, u8, sizeof (u8));
 
@@ -508,6 +499,7 @@ putcode_real (grub_uint32_t code, struct grub_term_output *term, int fixed_tab)
       .variant = 0,
       .attributes = 0,
       .ncomb = 0,
+      .combining = 0,
       .estimated_width = 1
     };
 
@@ -535,6 +527,7 @@ get_maxwidth (struct grub_term_output *term,
     .variant = 0,
     .attributes = 0,
     .ncomb = 0,
+    .combining = 0
   };
   return (grub_term_width (term)
 	  - grub_term_getcharwidth (term, &space_glyph) 
@@ -548,25 +541,13 @@ get_startwidth (struct grub_term_output *term,
   return ((term->getxy (term) >> 8) & 0xff) - margin_left;
 }
 
-static void
-fill_margin (struct grub_term_output *term, int r)
-{
-  int sp = (term->getwh (term) >> 8)
-    - (term->getxy (term) >> 8) - r;
-  if (sp > 0)
-    grub_print_spaces (term, sp);
-}
-
 static int
 print_ucs4_terminal (const grub_uint32_t * str,
 		     const grub_uint32_t * last_position,
 		     int margin_left, int margin_right,
 		     struct grub_term_output *term,
 		     struct term_state *state,
-		     int dry_run, int fixed_tab, unsigned skip_lines,
-		     unsigned max_lines,
-		     grub_uint32_t contchar,
-		     int primitive_wrap, int fill_right, struct grub_term_pos *pos)
+		     int dry_run, int fixed_tab)
 {
   const grub_uint32_t *ptr;
   grub_ssize_t startwidth = dry_run ? 0 : get_startwidth (term, margin_left);
@@ -575,59 +556,23 @@ print_ucs4_terminal (const grub_uint32_t * str,
   grub_ssize_t max_width = get_maxwidth (term, margin_left, margin_right);
   const grub_uint32_t *line_start = str, *last_space = str - 1;
   int lines = 0;
-  int i;
-  struct term_state local_state;
-
-  if (!state)
-    {
-      grub_memset (&local_state, 0, sizeof (local_state));
-      state = &local_state;
-    }
-
-  for (i = 0; i < state->bidi_stack_depth; i++)
-    putcode_real (state->bidi_stack[i] | (GRUB_UNICODE_LRE & ~0xff),
-		  term, fixed_tab);
 
   for (ptr = str; ptr < last_position; ptr++)
     {
       grub_ssize_t last_width = 0;
-
-      switch (*ptr)
-	{
-	case GRUB_UNICODE_LRE:
-	case GRUB_UNICODE_RLE:
-	case GRUB_UNICODE_LRO:
-	case GRUB_UNICODE_RLO:
-	  if (state->bidi_stack_depth >= (int) ARRAY_SIZE (state->bidi_stack))
-	    state->invalid_pushes++;
-	  else
-	    state->bidi_stack[state->bidi_stack_depth++] = *ptr;
-	  break;
-	case GRUB_UNICODE_PDF:
-	  if (state->invalid_pushes)
-	    state->invalid_pushes--;
-	  else if (state->bidi_stack_depth)
-	    state->bidi_stack_depth--;
-	  break;
-	}
       if (grub_unicode_get_comb_type (*ptr) == GRUB_UNICODE_COMB_NONE)
 	{
 	  struct grub_unicode_glyph c = {
 	    .variant = 0,
 	    .attributes = 0,
 	    .ncomb = 0,
+	    .combining = 0
 	  };
 	  c.base = *ptr;
-	  if (pos)
-	    {
-	      pos[ptr - str].x = line_width;
-	      pos[ptr - str].y = lines;
-	      pos[ptr - str].valid = 1;
-	    }
 	  line_width += last_width = grub_term_getcharwidth (term, &c);
 	}
 
-      if (*ptr == ' ' && !primitive_wrap)
+      if (*ptr == ' ')
 	{
 	  lastspacewidth = line_width;
 	  last_space = ptr;
@@ -636,13 +581,6 @@ print_ucs4_terminal (const grub_uint32_t * str,
       if (line_width > max_width || *ptr == '\n')
 	{
 	  const grub_uint32_t *ptr2;
-	  int wasn = (*ptr == '\n');
-
-	  if (wasn)
-	    {
-	      state->bidi_stack_depth = 0;
-	      state->invalid_pushes = 0;
-	    }
 
 	  if (line_width > max_width && last_space > line_start)
 	    ptr = last_space;
@@ -657,7 +595,7 @@ print_ucs4_terminal (const grub_uint32_t * str,
 
 	  lines++;
 
-	  if (!skip_lines && !dry_run)
+	  if (!dry_run)
 	    {
 	      for (ptr2 = line_start; ptr2 < ptr; ptr2++)
 		{
@@ -670,14 +608,9 @@ print_ucs4_terminal (const grub_uint32_t * str,
 		  putcode_real (*ptr2, term, fixed_tab);
 		}
 
-	      if (!wasn && contchar)
-		putcode_real (contchar, term, fixed_tab);
-	      if (fill_right)
-		fill_margin (term, margin_right);
-
-	      if (!contchar || max_lines != 1)
-		grub_putcode ('\n', term);
-	      if (state != &local_state && ++state->num_lines
+	      grub_print_spaces (term, margin_right);
+	      grub_putcode ('\n', term);
+	      if (state && ++state->num_lines
 		  >= (grub_ssize_t) grub_term_height (term) - 2)
 		{
 		  state->backlog_ucs4 = (ptr == last_space || *ptr == '\n') 
@@ -689,52 +622,19 @@ print_ucs4_terminal (const grub_uint32_t * str,
 	    }
 
 	  line_width -= lastspacewidth;
+	  if (!dry_run)
+	    grub_print_spaces (term, margin_left);
 	  if (ptr == last_space || *ptr == '\n')
 	    ptr++;
-	  else if (pos)
-	      {
-		pos[ptr - str].x = line_width - last_width;
-		pos[ptr - str].y = lines;
-		pos[ptr - str].valid = 1;
-	      }
-
 	  line_start = ptr;
-
-	  if (skip_lines)
-	    skip_lines--;
-	  else if (max_lines != (unsigned) -1)
-	    {
-	      max_lines--;
-	      if (!max_lines)
-		break;
-	    }
-	  if (!skip_lines && !dry_run)
-	    {
-	      if (!contchar)
-		grub_print_spaces (term, margin_left);
-	      else
-		grub_term_gotoxy (term, margin_left,
-				  grub_term_getxy (term) & 0xff);
-	      for (i = 0; i < state->bidi_stack_depth; i++)
-		putcode_real (state->bidi_stack[i] | (GRUB_UNICODE_LRE & ~0xff),
-			      term, fixed_tab);
-	    }
 	}
     }
 
-  if (pos)
-    {
-      pos[ptr - str].x = line_width;
-      pos[ptr - str].y = lines;
-      pos[ptr - str].valid = 1;
-    }
-
   if (line_start < last_position)
-      lines++;
-  if (!dry_run && !skip_lines && max_lines)
+    lines++;
+  if (!dry_run)
     {
       const grub_uint32_t *ptr2;
-
       for (ptr2 = line_start; ptr2 < last_position; ptr2++)
 	{
 	  /* Skip combining characters on non-UTF8 terminals.  */
@@ -745,9 +645,6 @@ print_ucs4_terminal (const grub_uint32_t * str,
 	    continue;
 	  putcode_real (*ptr2, term, fixed_tab);
 	}
-
-      if (fill_right)
-	fill_margin (term, margin_right);
     }
   return dry_run ? lines : 0;
 }
@@ -775,26 +672,20 @@ find_term_state (struct grub_term_output *term)
 }
 
 static int
-put_glyphs_terminal (struct grub_unicode_glyph *visual,
+put_glyphs_terminal (const struct grub_unicode_glyph *visual,
 		     grub_ssize_t visual_len,
 		     int margin_left, int margin_right,
 		     struct grub_term_output *term,
-		     struct term_state *state, int fixed_tab,
-		     grub_uint32_t contchar,
-		     int fill_right)
+		     struct term_state *state, int fixed_tab)
 {
-  struct grub_unicode_glyph *visual_ptr;
-  int since_last_nl = 1;
+  const struct grub_unicode_glyph *visual_ptr;
   for (visual_ptr = visual; visual_ptr < visual + visual_len; visual_ptr++)
     {
-      if (visual_ptr->base == '\n' && fill_right)
-	fill_margin (term, margin_right);
-
+      if (visual_ptr->base == '\n')
+	grub_print_spaces (term, margin_right);
       putglyph (visual_ptr, term, fixed_tab);
-      since_last_nl++;
       if (visual_ptr->base == '\n')
 	{
-	  since_last_nl = 0;
 	  if (state && ++state->num_lines
 	      >= (grub_ssize_t) grub_term_height (term) - 2)
 	    {
@@ -804,17 +695,10 @@ put_glyphs_terminal (struct grub_unicode_glyph *visual,
 	      return 1;
 	    }
 
-	  if (!contchar)
-	    grub_print_spaces (term, margin_left);
-	  else
-	    grub_term_gotoxy (term, margin_left,
-			      grub_term_getxy (term) & 0xff);
+	  grub_print_spaces (term, margin_left);
 	}
-      grub_unicode_destroy_glyph (visual_ptr);
+      grub_free (visual_ptr->combining);
     }
-  if (fill_right && since_last_nl)
-	fill_margin (term, margin_right);
-
   return 0;
 }
 
@@ -833,8 +717,7 @@ print_backlog (struct grub_term_output *term,
       ret = print_ucs4_terminal (state->backlog_ucs4,
 				 state->backlog_ucs4 + state->backlog_len,
 				 margin_left, margin_right, term, state, 0,
-				 state->backlog_fixed_tab, 0, -1, 0, 0,
-				 0, 0);
+				 state->backlog_fixed_tab);
       if (!ret)
 	{
 	  grub_free (state->free);
@@ -851,7 +734,7 @@ print_backlog (struct grub_term_output *term,
       ret = put_glyphs_terminal (state->backlog_glyphs,
 				 state->backlog_len,
 				 margin_left, margin_right, term, state,
-				 state->backlog_fixed_tab, 0, 0);
+				 state->backlog_fixed_tab);
       if (!ret)
 	{
 	  grub_free (state->free);
@@ -865,40 +748,22 @@ print_backlog (struct grub_term_output *term,
   return 0;
 }
 
-static grub_ssize_t
-getcharwidth (const struct grub_unicode_glyph *c, void *term)
-{
-  return grub_term_getcharwidth (term, c);
-}
-
 static int
 print_ucs4_real (const grub_uint32_t * str,
 		 const grub_uint32_t * last_position,
 		 int margin_left, int margin_right,
 		 struct grub_term_output *term, int backlog,
-		 int dry_run, int fixed_tab, unsigned skip_lines,
-		 unsigned max_lines,
-		 grub_uint32_t contchar, int fill_right,
-		 struct grub_term_pos *pos)
+		 int dry_run, int fixed_tab)
 {
   struct term_state *state = NULL;
 
   if (!dry_run)
     {
-      int xy;
       if (backlog)
 	state = find_term_state (term);
 
-      xy = term->getxy (term);
-      
-      if (((xy >> 8) & 0xff) < margin_left)
-	{
-	  if (!contchar)
-	    grub_print_spaces (term, margin_left - ((xy >> 8) & 0xff));
-	  else
-	    grub_term_gotoxy (term, margin_left,
-			      xy & 0xff);
-	}
+      if (((term->getxy (term) >> 8) & 0xff) < margin_left)
+	grub_print_spaces (term, margin_left - ((term->getxy (term) >> 8) & 0xff));
     }
 
   if ((term->flags & GRUB_TERM_CODE_TYPE_MASK) 
@@ -908,58 +773,41 @@ print_ucs4_real (const grub_uint32_t * str,
     {
       grub_ssize_t visual_len;
       struct grub_unicode_glyph *visual;
-      grub_ssize_t visual_len_show;
-      struct grub_unicode_glyph *visual_show;
       int ret;
-      struct grub_unicode_glyph *vptr;
+
+      auto grub_ssize_t getcharwidth (const struct grub_unicode_glyph *c);
+      grub_ssize_t getcharwidth (const struct grub_unicode_glyph *c)
+      {
+	return grub_term_getcharwidth (term, c);
+      }
 
       visual_len = grub_bidi_logical_to_visual (str, last_position - str,
-						&visual, getcharwidth, term,
+						&visual, getcharwidth,
 						get_maxwidth (term, 
 							      margin_left,
 							      margin_right),
-						dry_run ? 0 : get_startwidth (term, 
-									      margin_left),
-						contchar, pos, !!contchar);
+						get_startwidth (term, 
+								margin_left));
       if (visual_len < 0)
 	{
 	  grub_print_error ();
 	  return 0;
 	}
-      visual_show = visual;
-      for (; skip_lines && visual_show < visual + visual_len; visual_show++)
-	if (visual_show->base == '\n')
-	  skip_lines--;
-      if (max_lines != (unsigned) -1)
-	{
-	  for (vptr = visual_show;
-	       max_lines && vptr < visual + visual_len; vptr++)
-	    if (vptr->base == '\n')
-	      max_lines--;
-
-	  visual_len_show = vptr - visual_show;	  
-	}
-      else
-	visual_len_show = visual + visual_len - visual_show;
-
       if (dry_run)
 	{
+	  struct grub_unicode_glyph *vptr;
 	  ret = 0;
-	  for (vptr = visual_show; vptr < visual_show + visual_len_show; vptr++)
+	  for (vptr = visual; vptr < visual + visual_len; vptr++)
 	    if (vptr->base == '\n')
 	      ret++;
-	  if (visual_len_show && visual[visual_len_show - 1].base != '\n')
+	  if (visual_len && visual[visual_len - 1].base != '\n')
 	    ret++;
-	  for (vptr = visual; vptr < visual + visual_len; vptr++)
-	    grub_unicode_destroy_glyph (vptr);
 	  grub_free (visual);
 	}
       else
 	{
-	  ret = put_glyphs_terminal (visual_show, visual_len_show, margin_left,
-				     margin_right,
-				     term, state, fixed_tab, contchar, fill_right);
-
+	  ret = put_glyphs_terminal (visual, visual_len, margin_left,
+				     margin_right, term, state, fixed_tab);
 	  if (!ret)
 	    grub_free (visual);
 	  else
@@ -968,22 +816,7 @@ print_ucs4_real (const grub_uint32_t * str,
       return ret;
     }
   return print_ucs4_terminal (str, last_position, margin_left, margin_right,
-			      term, state, dry_run, fixed_tab, skip_lines,
-			      max_lines, contchar, !!contchar, fill_right, pos);
-}
-
-void
-grub_print_ucs4_menu (const grub_uint32_t * str,
-		      const grub_uint32_t * last_position,
-		      int margin_left, int margin_right,
-		      struct grub_term_output *term,
-		      int skip_lines, int max_lines,
-		      grub_uint32_t contchar,
-		      struct grub_term_pos *pos)
-{
-  print_ucs4_real (str, last_position, margin_left, margin_right,
-		   term, 0, 0, 1, skip_lines, max_lines,
-		   contchar, 1, pos);
+			      term, state, dry_run, fixed_tab);
 }
 
 void
@@ -993,7 +826,7 @@ grub_print_ucs4 (const grub_uint32_t * str,
 		 struct grub_term_output *term)
 {
   print_ucs4_real (str, last_position, margin_left, margin_right,
-		   term, 0, 0, 1, 0, -1, 0, 0, 0);
+		   term, 0, 0, 1);
 }
 
 int
@@ -1003,31 +836,31 @@ grub_ucs4_count_lines (const grub_uint32_t * str,
 		       struct grub_term_output *term)
 {
   return print_ucs4_real (str, last_position, margin_left, margin_right,
-			  term, 0, 1, 1, 0, -1, 0, 0, 0);
+			  term, 0, 1, 1);
 }
 
 void
-grub_xnputs (const char *str, grub_size_t msg_len)
+grub_xputs_normal (const char *str)
 {
   grub_uint32_t *unicode_str = NULL, *unicode_last_position;
   int backlog = 0;
   grub_term_output_t term;
 
   grub_error_push ();
-
-  unicode_str = grub_malloc (msg_len * sizeof (grub_uint32_t));
- 
+  grub_utf8_to_ucs4_alloc (str, &unicode_str,
+  			   &unicode_last_position);
   grub_error_pop ();
 
   if (!unicode_str)
     {
-      for (; msg_len--; str++, msg_len++)
+      for (; *str; str++)
 	{
 	  struct grub_unicode_glyph c =
 	    {
 	      .variant = 0,
 	      .attributes = 0,
 	      .ncomb = 0,
+	      .combining = 0,
 	      .estimated_width = 1,
 	      .base = *str
 	    };
@@ -1049,15 +882,11 @@ grub_xnputs (const char *str, grub_size_t msg_len)
       return;
     }
 
-  msg_len = grub_utf8_to_ucs4 (unicode_str, msg_len,
-			       (grub_uint8_t *) str, -1, 0);
-  unicode_last_position = unicode_str + msg_len;
-
   FOR_ACTIVE_TERM_OUTPUTS(term)
   {
     int cur;
     cur = print_ucs4_real (unicode_str, unicode_last_position, 0, 0,
-			   term, grub_more, 0, 0, 0, -1, 0, 0, 0);
+			   term, grub_more, 0, 0);
     if (cur)
       backlog = 1;
   }
@@ -1076,12 +905,6 @@ grub_xnputs (const char *str, grub_size_t msg_len)
   grub_free (unicode_str);
 }
 
-void
-grub_xputs_normal (const char *str)
-{
-  grub_xnputs (str, grub_strlen (str));
-}
-
 void
 grub_cls (void)
 {
